@@ -71,10 +71,8 @@ rtr_index_build_node_ptr(
 					pointer */
 	ulint			page_no,/*!< in: page number to put in node
 					pointer */
-	mem_heap_t*		heap,	/*!< in: memory heap where pointer
+	mem_heap_t*		heap);	/*!< in: memory heap where pointer
 					created */
-	ulint			level);	/*!< in: level of rec in tree:
-					0 means leaf level */
 
 /*************************************************************//**
 Splits an R-tree index page to halves and inserts the tuple. It is assumed
@@ -153,7 +151,6 @@ rtr_rec_cal_increase(
 				dtuple in some of the common fields, or which
 				has an equal number or more fields than
 				dtuple */
-	const rec_offs*	offsets,/*!< in: array returned by rec_get_offsets() */
 	double*		area);	/*!< out: increased area */
 
 /****************************************************************//**
@@ -163,7 +160,6 @@ dberr_t
 rtr_ins_enlarge_mbr(
 /*=================*/
 	btr_cur_t*		cursor,	/*!< in: btr cursor */
-	que_thr_t*		thr,	/*!< in: query thread */
 	mtr_t*			mtr);	/*!< in: mtr */
 
 /**************************************************************//**
@@ -173,10 +169,10 @@ void
 rtr_non_leaf_stack_push(
 /*====================*/
 	rtr_node_path_t*	path,		/*!< in/out: search path */
-	ulint			pageno,		/*!< in: pageno to insert */
+	uint32_t		pageno,		/*!< in: pageno to insert */
 	node_seq_t		seq_no,		/*!< in: Node sequence num */
 	ulint			level,		/*!< in: index level */
-	ulint			child_no,	/*!< in: child page no */
+	uint32_t		child_no,	/*!< in: child page no */
 	btr_pcur_t*		cursor,		/*!< in: position cursor */
 	double			mbr_inc);	/*!< in: MBR needs to be
 						enlarged */
@@ -194,23 +190,8 @@ rtr_non_leaf_insert_stack_push(
 	double			mbr_inc);	/*!< in: MBR needs to be
 						enlarged */
 
-/*****************************************************************//**
-Allocates a new Split Sequence Number.
-@return new SSN id */
-UNIV_INLINE
-node_seq_t
-rtr_get_new_ssn_id(
-/*===============*/
-	dict_index_t*	index);		/*!< in: the index struct */
-
-/*****************************************************************//**
-Get the current Split Sequence Number.
-@return current SSN id */
-UNIV_INLINE
-node_seq_t
-rtr_get_current_ssn_id(
-/*===================*/
-	dict_index_t*	index);		/*!< in/out: the index struct */
+#define rtr_get_new_ssn_id(index) (index)->assign_ssn()
+#define rtr_get_current_ssn_id(index) (index)->ssn()
 
 /********************************************************************//**
 Create a RTree search info structure */
@@ -406,9 +387,6 @@ rtr_merge_and_update_mbr(
 	rec_offs*		offsets,	/*!< in: rec offsets */
 	rec_offs*		offsets2,	/*!< in: rec offsets */
 	page_t*			child_page,	/*!< in: the child page. */
-	buf_block_t*		merge_block,	/*!< in: page to merge */
-	buf_block_t*		block,		/*!< in: page be merged */
-	dict_index_t*		index,		/*!< in: index */
 	mtr_t*			mtr);		/*!< in: mtr */
 
 /*************************************************************//**
@@ -416,10 +394,8 @@ Deletes on the upper level the node pointer to a page. */
 void
 rtr_node_ptr_delete(
 /*================*/
-	dict_index_t*	index,	/*!< in: index tree */
-	btr_cur_t*	sea_cur,/*!< in: search cursor, contains information
+	btr_cur_t*	cursor,	/*!< in: search cursor, contains information
 				about parent nodes in search */
-	buf_block_t*	block,	/*!< in: page whose node pointer is deleted */
 	mtr_t*		mtr);	/*!< in: mtr */
 
 /****************************************************************//**
@@ -431,10 +407,7 @@ rtr_merge_mbr_changed(
 	btr_cur_t*	cursor2,	/*!< in: the other cursor */
 	rec_offs*	offsets,	/*!< in: rec offsets */
 	rec_offs*	offsets2,	/*!< in: rec offsets */
-	rtr_mbr_t*	new_mbr,	/*!< out: MBR to update */
-	buf_block_t*	merge_block,	/*!< in: page to merge */
-	buf_block_t*	block,		/*!< in: page be merged */
-	dict_index_t*	index);		/*!< in: index */
+	rtr_mbr_t*	new_mbr);	/*!< out: MBR to update */
 
 
 /**************************************************************//**
@@ -511,7 +484,7 @@ rtr_info_reinit_in_cursor(
 @param[in]	tuple	range tuple containing mbr, may also be empty tuple
 @param[in]	mode	search mode
 @return estimated number of rows */
-int64_t
+ha_rows
 rtr_estimate_n_rows_in_range(
 	dict_index_t*	index,
 	const dtuple_t*	tuple,
